import contextvars

from typing import Any, Callable, Coroutine


class StreamingContext():
    from langchain.callbacks.base import AsyncCallbackHandler

    class StreamingContextCallback(AsyncCallbackHandler):

        async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
            await StreamingContext.get_context().on_new_token(token)

        async def on_llm_end(self, response, *args, **kwargs):
            if StreamingContext.get_context().stream_to_stdout:
                print()

    context_var = contextvars.ContextVar('streaming_context')

    def __init__(self, callback: Callable[[str], None] = None, callback_async: Callable[[str], Coroutine[None, None, None]]=None, stream_to_stdout: bool = False) -> None:
        self.callback = callback
        self.callback_async = callback_async
        self.stream_to_stdout = stream_to_stdout
        self.token_colors = ['\033[90m', '\033[0m']

    def __enter__(self):
        self.__class__.context_var.set(self)

    @classmethod
    def get_context(cls) -> 'StreamingContext':
        return cls.context_var.get(None)

    async def on_new_token(self, token: str):
        if token:
            if self.callback:
                self.callback(token)
            if self.callback_async:
                await self.callback_async(token)
            if self.stream_to_stdout:
                reset_color = '\033[0m'
                if token.strip() and token!="\n":
                    current_color = self.token_colors[0]
                    self.token_colors.reverse()
                else:
                    current_color='\033[40m'
                print('{}{}{}'.format(current_color, token if token !=
                    "" else '\u2022', reset_color), end='')

    def __exit__(self, exc_type, exc_value, traceback):
        self.context_var.set(None)




import re
import inspect
import json
import logging
import os
from textwrap import dedent
import yaml
from enum import Enum
from typing import Any, Coroutine, Dict, List, Union, Optional, Tuple, get_origin
from pydantic import BaseConfig, BaseModel, Extra, Field
from pydantic.fields import ModelField
from langchain.llms.base import BaseLanguageModel
from langchain.chat_models import ChatOpenAI
from langchain.schema import BaseMessage
from typing_inspect import is_generic_type, is_union_type




class LlmSelector(BaseModel):

    rules:List[dict]=[]
    llms:Dict[int,BaseLanguageModel]=Field(default_factory=dict)
    streamable_llms_cache:Dict[int,BaseLanguageModel]=Field(default_factory=dict)
    generation_min_tokens:Optional[int]
    prompt_to_generation_ratio:Optional[float]

    def __init__(self, generation_min_tokens:int=None, prompt_to_generation_ratio:float=1/3):
        """ Create a LlmSelector that will select the llm based on the length of the prompt.
        
        Args:
            generation_min_tokens (int, optional): The minimum number of tokens that the llm is expecting generate. Defaults to None (prompt_to_generation_ratio will be used).
            prompt_to_generation_ratio (float, optional): The ratio of the prompt length to the generation length. Defaults to 1/3. 
        """
        super().__init__(generation_min_tokens=generation_min_tokens, prompt_to_generation_ratio=prompt_to_generation_ratio)


    def with_llm(self, llm:BaseLanguageModel):
        """ this will automatically add a rule with token window based on the model name. Only works for OpenAI and Anthropic models."""
        max_tokens = self.get_model_window(llm.model_name)
        if max_tokens:
            self.with_llm_rule(llm, max_tokens)
        else:
            raise Exception(f"Could not find a token limit for model {llm.model_name}. Please use `with_llm_rule` and specify the max_tokens for your model.")
 
        return self

    def with_llm_rule(self, llm:BaseLanguageModel, max_tokens:int):
        i=-1
        for i,rule in enumerate(self.rules):
            rule_max_token=rule.get("max_tokens")
            if rule_max_token and rule_max_token > max_tokens:
                i=i-1
                break
        i=i+1
        self.rules.insert(i, dict(max_tokens=max_tokens))
        self.llms[i]=llm    
        return self
        
    def get_model_window(self, model_name:str)->int:
        for model_pattern, max_tokens in MODEL_LIMITS.items():
            if re.match(model_pattern, model_name):
                return max_tokens
        

    def get_llm(self, prompt:Union[str,List[BaseMessage]], function_schemas:List[dict]=None, expected_generated_tokens:int=None, streaming=False)->BaseLanguageModel:
        """Picks the best LLM based on the rules and the prompt length.

        Args:
            prompt (Union[str,List[BaseMessage]]): the prompt ... messages or string
            function_schemas (List[dict], optional): openAI function schemas. Defaults to None. (are included in the token limit)
            expected_generated_tokens (int, optional): Number of tokens we expect model to generate. Help for better precision. If None, the prompt_to_generation_ratio will be used (defaults to 1/3 - means 30% above the prompt length)

        """
        if not self.llms:
            raise Exception("No LLMs rules added to the LlmSelector")
        
        
        first_rule = self.rules[0]
        first_token_threshold = first_rule.get("max_tokens")
        total_tokens_estimate = self.get_expected_total_tokens(prompt, function_schemas=function_schemas, estimate=True, expected_generated_tokens=expected_generated_tokens)
        if total_tokens_estimate<first_token_threshold:
            result_index = 0
        else:
            total_tokens = self.get_expected_total_tokens(prompt, function_schemas=function_schemas, estimate=False, expected_generated_tokens=expected_generated_tokens) 
            for i, rule in enumerate(self.rules):
                max_tokens = rule.get("max_tokens")
                if max_tokens and max_tokens >=total_tokens:
                    result_index = i
                    break

            # if no condition is met, return the last llm
            result_index = i
        print_log(f"LLMSelector: Using {'default' if result_index==0 else str(result_index)+'-th'} LLM: {getattr(self.llms[result_index],'model_name', self.llms[result_index].__class__.__name__)}", logging.DEBUG )
        if streaming:
            if result_index not in self.streamable_llms_cache:
                self.streamable_llms_cache[result_index] = make_llm_streamable(self.llms[result_index])
            return self.streamable_llms_cache[result_index]
        else:
            return self.llms[result_index]
    
    def get_expected_total_tokens(self, prompt:Union[str,List[BaseMessage]], function_schemas:List[dict]=None, estimate:bool=True,expected_generated_tokens=None)->int:
        expected_generated_tokens = expected_generated_tokens or self.generation_min_tokens or 0
        prompt_tokens = self.get_token_count(prompt, function_schemas=function_schemas, estimate=estimate) 
        if expected_generated_tokens:
            return prompt_tokens + expected_generated_tokens
        else:
             return prompt_tokens * (1+(self.prompt_to_generation_ratio or 0))
        
    
    def get_token_count(self, prompt:Union[str,List[BaseMessage]], function_schemas:List[dict]=None, estimate:bool=True)->int:
        """Get the number of tokens in the prompt. If estimate is True, it will use a fast estimation, otherwise it will use the llm to count the tokens (slower)"""
        if estimate:
            num_tokens = int(len(prompt)/2)
        else:
            num_tokens = count_tokens(prompt, llm=self.llms[0] ) # note: we will use the first llm to count the tokens... it should be the same general type, and if not, it's ok, should be close enough
        
        if function_schemas:
            num_tokens += self.get_token_count(json.dumps(function_schemas), estimate=estimate)
        return num_tokens
    




class GlobalSettings(BaseModel):
    default_llm: Optional[BaseLanguageModel] = None
    default_streaming_llm: Optional[BaseLanguageModel] = None
    logging_level: int = logging.INFO
    verbose: bool = False
    llm_selector: Optional[LlmSelector] = None

    class Config:
        allow_population_by_field_name = True
        extra = Extra.allow

    @classmethod
    def define_settings(cls,
                        settings_type="default",
                        default_llm:BaseLanguageModel=None,
                        default_streaming_llm:BaseLanguageModel=None,
                        logging_level=logging.INFO,
                        verbose=None,
                        llm_selector: Optional["LlmSelector"] = None,
                        **kwargs
                        ):
        """ Define the global settings for the project.
        
        Args:
            settings_type (str, optional): The name of the settings. Defaults to "default".
            default_llm (BaseLanguageModel, optional): The default language model to use. Defaults to None.
            default_streaming_llm (BaseLanguageModel, optional): The default streaming language model to use. Defaults to None.
            llm_selector (Optional[LlmSelector], optional): The language model selector to use. Defaults to None.
            logging_level (int, optional): The logging level to use. Defaults to logging.INFO.

        """
        if llm_selector is None and default_llm is None and default_streaming_llm is None:
            # only use llm_selector if no default_llm and default_streaming_llm is defined, because than we dont know what rules to set up
            default_llm = ChatOpenAI(temperature=0.0, model="gpt-3.5-turbo-0613") #  '-0613' - has function calling
            default_streaming_llm = make_llm_streamable(default_llm)
            llm_selector = LlmSelector()\
                .with_llm(default_llm)\
                .with_llm(ChatOpenAI(temperature=0.0, model="gpt-3.5-turbo-16k-0613"))  #  '-0613' - has function calling
        
        else:
            if default_llm is None:
                default_llm = ChatOpenAI(temperature=0.0, model="gpt-3.5-turbo-0613")  #  '-0613' - has function calling
            if default_streaming_llm is None:
                default_streaming_llm = make_llm_streamable(default_llm)
            
        if verbose is None:
            verbose = os.environ.get("LANGCHAIN_DECORATORS_VERBOSE", False) in [True,"true","True","1"]
        settings = cls(default_llm=default_llm, default_streaming_llm=default_streaming_llm,
                       logging_level=logging_level, verbose=verbose, llm_selector=llm_selector, **kwargs)
        if not hasattr(GlobalSettings, "registry"):
            setattr(GlobalSettings, "registry", {})
        GlobalSettings.registry[settings_type] = settings

    @classmethod
    def get_current_settings(cls) -> "GlobalSettings":
        if not hasattr(GlobalSettings, "settings_type"):
            setattr(GlobalSettings, "settings_type", "default")
        if not hasattr(GlobalSettings, "registry"):
            GlobalSettings.define_settings()
        return GlobalSettings.registry[GlobalSettings.settings_type]

    @classmethod
    def switch_settings(cls, project_name):
        GlobalSettings.settings_type = project_name




class LogColors(Enum):
    WHITE_BOLD = "\033[1m"
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    DARK_GRAY = '\033[90m'
    WHITE = '\033[39m'
    BLACK_AND_WHITE = '\033[40m'

    # Define some reset codes to restore the default text color
    RESET = '\033[0m'


def print_log(log_object: Any, log_level: int, color: LogColors = None):
    settings = GlobalSettings.get_current_settings()
    if settings.logging_level <= log_level or settings.verbose:
        if isinstance(log_object, str):
            pass
        elif isinstance(log_object, dict):
            log_object = yaml.safe_dump(log_object)
        elif isinstance(log_object, BaseModel):
            log_object = yaml.safe_dump(log_object.dict())

        if color is None:
            if log_level >= logging.ERROR:
                color = LogColors.RED
            elif log_level >= logging.WARNING:
                color = LogColors.YELLOW
            elif log_level >= logging.INFO:
                color = LogColors.GREEN
            else:
                color = LogColors.DARK_GRAY
        if type(color) is LogColors:
            color = color.value
        reset = LogColors.RESET.value if color else ""
        print(f"{color}{log_object}{reset}\n", flush=True)


class PromptTypeSettings:
    def __init__(self, llm: BaseLanguageModel = None,  color: LogColors = None, log_level: Union[int, str] = "info", capture_stream: bool = False, llm_selector: "LlmSelector" = None):
        self.color = color
        if isinstance(log_level, str):
            log_level = getattr(logging, log_level.upper())
        self.log_level = log_level
        self.capture_stream = capture_stream
        self.llm = llm
        self.llm_selector = llm_selector
    

    def as_verbose(self):
        return PromptTypeSettings(llm=self.llm, color=self.color, log_level=100, capture_stream=self.capture_stream)


class PromptTypes:
    UNDEFINED: PromptTypeSettings = PromptTypeSettings(
        color=LogColors.BLACK_AND_WHITE, log_level=logging.DEBUG)
    BIG_CONTEXT: PromptTypeSettings = PromptTypeSettings(
        llm=ChatOpenAI(temperature=0.0, model="gpt-3.5-turbo-16k"), 
        color=LogColors.BLACK_AND_WHITE, log_level=logging.DEBUG)
        
    AGENT_REASONING: PromptTypeSettings = PromptTypeSettings(
        color=LogColors.GREEN, log_level=logging.INFO)
    TOOL: PromptTypeSettings = PromptTypeSettings(
        color=LogColors.BLUE, log_level=logging.INFO)
    FINAL_OUTPUT: PromptTypeSettings = PromptTypeSettings(
        color=LogColors.YELLOW, log_level=logging.INFO)


def get_func_return_type(func: callable)->Tuple:
    return_type = func.__annotations__.get("return",None)
    if inspect.iscoroutinefunction(func):
        if return_type:
            if is_generic_type(return_type):
                return_type_origin = get_origin(return_type)
                if return_type_origin and issubclass(return_type_origin, Coroutine):
                    return_type_args = getattr(return_type, '__args__', None)
                    if return_type_args and len(return_type_args) == 3:
                        return_type = return_type_args[2]
                    else:
                        raise Exception(f"Invalid Coroutine annotation {return_type}. Expected Coroutine[ any , any, <return_type>] or just <return_type>")
                else:
                    return_type = return_type_origin
        else:
            
            if return_type and issubclass(return_type, Coroutine):
                return None
            else:
                return_type = return_type
    
    if return_type and is_union_type(return_type):
        return_type_args = getattr(return_type, '__args__', None)
        if return_type_args and len(return_type_args) == 2 and return_type_args[1] == type(None):
            return return_type_args[0]
        else:
            raise Exception(f"Invalid Union annotation {return_type}. Expected Union[ <return_type>, None] or just <return_type>")
    elif is_generic_type(return_type):
        # this should cover list and dict
        return get_origin(return_type)
    else:
        return return_type
            
            
def get_function_docs(func: callable)->Tuple:
    if not func.__doc__:
        return None
    fist_line, rest = func.__doc__.split('\n', 1) if '\n' in func.__doc__ else (func.__doc__, "")
    # we dedent the first line separately,because its common that it often starts right after """
    fist_line = fist_line.strip()
    if fist_line:
        fist_line+="\n"
    docs = fist_line + dedent(rest)
    return docs
    

            
def get_function_full_name(func: callable)->str:
    return  f"{func.__module__}.{func.__name__}" if not func.__module__=="__main__" else func.__name__
    


def get_arguments_as_pydantic_fields(func) -> Dict[str, ModelField]:
    argument_types = {}
    model_config = BaseConfig()
    for arg_name, arg_desc in inspect.signature(func).parameters.items():
        if arg_name != "self":
            default = arg_desc.default if arg_desc.default!=inspect.Parameter.empty else None
            if arg_desc.annotation==inspect._empty:
                raise Exception(f"Argument '{arg_name}' of function {func.__name__} has no type annotation")
            argument_types[arg_name] = ModelField(
                class_validators=None,
                model_config=model_config,
                name=arg_name, 
                type_=arg_desc.annotation,
                default=default,
                required= arg_desc.default==inspect.Parameter.empty
                )
            
    return argument_types


def make_llm_streamable(llm:BaseLanguageModel):
    try:
        if hasattr(llm,"lc_kwargs"):
            # older version support
            lc_kwargs = {**llm.lc_kwargs}
        else:
            lc_kwargs = {
                k: getattr(llm, k, v)
                for k, v in llm._lc_kwargs.items()
                if not (llm.__exclude_fields__ or {}).get(k, False)  # type: ignore
            }
        lc_kwargs["streaming"] = True
        return llm.__class__(**lc_kwargs)
    except Exception as e:
        logging.warning(f"Could not make llm {llm} streamable. Error: {e}")
        

def count_tokens(prompt: Union[str,List[BaseMessage]], llm:BaseLanguageModel) -> int:
    """Returns the number of tokens in a text string."""
    if isinstance(prompt,str):
        return llm.get_num_tokens(prompt)
    elif isinstance(prompt,list):
        return llm.get_num_tokens_from_messages(prompt)


MODEL_LIMITS={
    "gpt-3.5-turbo-16k.*": 16_384,
    "gpt-3.5-turbo.*": 4_096,

    "text-davinci-003.*": 4_097,
    "text-davinci-003.*": 4_097,
    "code-davinci-002.*": 8_001,

    "gpt-4-32k.*": 32_768,
    "gpt-4.*": 8_192,
    
    "claude-v1":9000,
    "claude-v1":9000,
    r"claude-v\d(\.\d+)?-100k":100_000,
}




import logging
import inspect

from functools import wraps
from textwrap import dedent
from typing import Callable, List, Optional,  Union


from langchain import LLMChain,  PromptTemplate
from langchain.tools.base import BaseTool
from langchain.schema import BaseOutputParser
from langchain.llms.base import BaseLanguageModel

from promptwatch import register_prompt_template

from .schema import OutputWithFunctionCall

from .chains import LLMDecoratorChainWithFunctionSupport, LLMDecoratorChain



from .common import *
from .prompt_template import PromptDecoratorTemplate
from .output_parsers import *
from .streaming_context import StreamingContext


def llm_prompt(
        prompt_type:PromptTypeSettings=PromptTypes.UNDEFINED, # do not change the order of this first parameter unless you will change also the fist few lines... since we are handling cases when decorator is used with and without arguments too, than this will be the func
        template_format:str = "f-string-extra",
        output_parser:Union[str,None, BaseOutputParser]="auto", 
        stop_tokens:List[str]=None, 
        template_name:str=None, 
        template_version:str=None, 
        capture_stream:bool=None,
        llm:Optional[BaseLanguageModel]=None,
        format_instructions_parameter_key:str="FORMAT_INSTRUCTIONS",
        retry_on_output_parsing_error:bool=True,
        verbose:bool=None,
        expected_gen_tokens:Optional[int]=None,
        ):
    """
    Decorator for functions that turns a regular function into a LLM prompt executed with default model and settings.
    
    This can be applied on any function that has a docstring with a prompt template. 
    If the function is async, the prompt will be executed asynchronously (with all the langchain async infrastructure).

    Note that the code of the function will never be executed... 

    Args:
        `prompt_type`: (Optional[PromptTypeSettings]) - This allows you mark your prompt with one of the predefined prompt types (see PromptTypes class - but you can subclass it!) to predefine some settings like LLM or style and color of logging into console.

        `template_format` (Optional[str]): one of [ `f-string` | `f-string-extra` ] ... f-string-extra is a superset of f-string template formats, enabling for optional sections.

        `output_parser` (Optional[str]): one of [ `auto` | `json` | `str` | `list` ] or `None` or langchain OutputParser object - you can control how will the output be parsed. 
        
            `auto` - default - determine the output type automatically based on output type annotations

            `str` or `None` - will return plain string output

            `list` - will parse bullet or numbered list (each item on a new line) as a list

            `boolean` - will parse the output as boolean. Expects clear Yes/No in the output

            `json` - will parse the output as json

            `functions` - will use the OpenAI functions to generate the output in desired format ... only for pydataic models and ChatOpenAI model

            `markdown` - will parse the output as markdown sections, the name of each section will be returned as a key and the content as a value. For nested sections, the value will be a dict with the same structure.

            `pydantic` - will parse the output as json and then convert into a pydantic model


        `stop_tokens` (Optional[List[str]]): list of stop tokens to instruct the LLM to stop generating text when it encounters any of these tokens. If not provided, the default stop tokens of the LLM will be used.

        `format_instructions_parameter_key` - name of the format instructions parameter - this will enable you to include the instructions on how LLM should format the output, generated by the output_parsers 
        ... if you include this into your prompt (docs), you don't need to reinvent the formatting instructions. 
        This works pretty well if you have an annotated pydantic model as an function output. If you are expecting a dict, you should probably include your own formatting instructions, since there is not much to infer from a dict structure.

        `retry_on_output_parsing_error` - whether to try to re-format the output if the output parser fails to parse the output by another LLM call

        `verbose` - whether to print the response from LLM into console

        `expected_gen_tokens` - hint for LLM selector ... if not set, default values of the LLM selector will be used (usually 1/3 of the prompt length)
    """
    


    if  callable(prompt_type):
        # this is the case when the decorator is called without arguments
        # we initialize params with default values
        func = prompt_type
        prompt_type = PromptTypes.UNDEFINED
    else:
        func = None
    
    if verbose is None:
        verbose = GlobalSettings.get_current_settings().verbose
    
    if verbose:
        if prompt_type:
            prompt_type = prompt_type.as_verbose()
        else:
            prompt_type = PromptTypeSettings(color=LogColors.DARK_GRAY,log_level=100, capture_stream=capture_stream)
            
    
    
    def decorator(func):
        prompt_str = dedent(func.__doc__)
        name=func.__name__
        full_name=f"{func.__module__}.{name}" if func.__module__!="__main__" else name
        is_async = inspect.iscoroutinefunction(func)


        

        if prompt_type:
            _capture_stream = prompt_type.capture_stream if capture_stream is None else capture_stream
        else:
            _capture_stream = capture_stream
        if _capture_stream and not is_async:
            print_log(f"Warning: capture_stream=True is only supported for async functions. Ignoring capture_stream for {full_name}", logging.WARNING, LogColors.YELLOW)
            _capture_stream=False
            
                
                
       

        def prepare_call_args(*args, **kwargs):
            global_settings = GlobalSettings.get_current_settings()

            capture_stream=_capture_stream

            if "capture_stream" in kwargs:
                if not isinstance(capture_stream,bool):
                    raise ValueError("capture_stream is a reserved kwarg and must be of type bool")
                capture_stream=kwargs["capture_stream"]
                del kwargs["capture_stream"]

            if capture_stream and not StreamingContext.get_context():
                print_log(f"INFO: Not inside StreamingContext. Ignoring capture_stream for {full_name}", logging.DEBUG, LogColors.WHITE)
                capture_stream=False
            
            
            if not llm:
                if prompt_type and prompt_type.llm_selector:
                    llm_selector= prompt_type.llm_selector
                else:
                    llm_selector=  global_settings.llm_selector 

                if capture_stream and not llm_selector:
                    if not global_settings.default_streaming_llm:
                        print_log(f"Warning: capture_stream on {name} is on, but the default LLM {llm} doesn't seem to be supporting streaming.", logging.WARNING, LogColors.YELLOW)
                        
                    prompt_llm=global_settings.default_streaming_llm or global_settings.default_llm
                else:
                    prompt_llm = global_settings.default_llm

                
            else:
                prompt_llm=llm
                llm_selector=None # if LLM is explicitly provided, we don't use the selector
                if capture_stream:
                    if  hasattr(llm,"streaming"):
                        if not getattr(llm, "streaming"):
                            print_log(f"Warning: capture_stream on {name} is on, but the provided LLM {llm} doesn't have streaming on! Stream wont be captured", logging.WARNING, LogColors.YELLOW)
                    else:
                        print_log(f"Warning: capture_stream on {name} is on, but the provided LLM {llm} doesn't seem to be supporting streaming.", logging.WARNING, LogColors.YELLOW)
                   
                

            input_variables_source=None
            if len(args)==1 and hasattr(args[0],"__dict__"):
                # is a proper object
                input_variables_source = args[0]

            elif len(args)>1:
                raise Exception(f"Positional arguments are not supported for prompt functions. Only one positional argument as an object with attributes as a source of inputs is supported. Got: {args}")
            
            
            prompt_template = PromptDecoratorTemplate.from_func(func, 
                                                            template_format=template_format, 
                                                            output_parser=output_parser, 
                                                            format_instructions_parameter_key=format_instructions_parameter_key,
                                                            template_name=template_name,
                                                            template_version=template_version,
                                                            prompt_type=prompt_type,
                                                            )
            if prompt_template.default_values:
                kwargs = {**prompt_template.default_values, **kwargs}

            if "callbacks" in kwargs:
                callbacks=kwargs.pop("callbacks")
            else:
                callbacks=[]
            
            if capture_stream:
                callbacks.append(StreamingContext.StreamingContextCallback())
            

            if "memory" in kwargs:
                memory = kwargs.pop("memory")
            else:
                memory=None

            if "functions" in kwargs:
                functions=kwargs.pop("functions")
            else:
                functions=None

                
            if functions:
                llmChain = LLMDecoratorChainWithFunctionSupport(llm=prompt_llm, prompt=prompt_template,  memory=memory, functions=functions, llm_selector=llm_selector, capture_stream=capture_stream, expected_gen_tokens=expected_gen_tokens)
            elif isinstance(prompt_template.output_parser, OpenAIFunctionsPydanticOutputParser):
                function=prompt_template.output_parser.build_llm_function()
                kwargs["function_call"] = function
                llmChain = LLMDecoratorChainWithFunctionSupport(llm=prompt_llm, prompt=prompt_template,  memory=memory, functions=[function], llm_selector=llm_selector, capture_stream=capture_stream, expected_gen_tokens=expected_gen_tokens )
            else:
                llmChain = LLMDecoratorChain(llm=prompt_llm, prompt=prompt_template,  memory=memory, llm_selector=llm_selector, capture_stream=capture_stream, expected_gen_tokens=expected_gen_tokens)
            other_supported_kwargs={"stop","callbacks","function_call"}
            unexpected_inputs = [key for key in kwargs if key not in prompt_template.input_variables and key not in other_supported_kwargs ]
            if unexpected_inputs:
                raise TypeError(f"Unexpected inputs for prompt function {full_name}: {unexpected_inputs}. \nValid inputs are: {prompt_template.input_variables}\nHint: Make sure that you've used all the inputs in the template")
            
            missing_inputs = [key for key in prompt_template.input_variables if key not in kwargs ]
            if format_instructions_parameter_key in missing_inputs:
                missing_inputs.remove(format_instructions_parameter_key)
                kwargs[format_instructions_parameter_key]=None #init the format instructions with None... will be filled later
            if memory and memory.memory_key in missing_inputs:
                missing_inputs.remove(memory.memory_key)

            if missing_inputs:
                if input_variables_source:
                    missing_value={}
                    for key in missing_inputs:
                        value= getattr(input_variables_source, key,missing_value)
                        if value is missing_value:
                            raise TypeError(f"Missing a input for prompt function {full_name}: {key}.")
                        else:
                            kwargs[key] = value


                        
                else:
                    raise TypeError(f"{full_name}: missing 1 required keyword-only argument: {missing_inputs}")
                
            
            if stop_tokens:
                kwargs["stop"]=stop_tokens
            call_args = {"inputs":kwargs, "return_only_outputs":True, "callbacks":callbacks}
           
            return llmChain, call_args
        
        
        
        def get_retry_parse_call_args(prompt_template:PromptDecoratorTemplate, exception:OutputParserExceptionWithOriginal, get_original_prompt:Callable):
            logging.warning(msg=f"Failed to parse output for {full_name}: {exception}\nRetrying...")
            if format_instructions_parameter_key not in prompt_str:
                logging.warning(f"Please note that we didn't find a {format_instructions_parameter_key} parameter in the prompt string. If you don't include it in your prompt template, you need to provide your custom formatting instructions.")    
            if exception.original_prompt_needed_on_retry:
                original_prompt=get_original_prompt()
            else:
                original_prompt=""
            retry_parse_template = PromptTemplate.from_template("{original_prompt}This is our original response {original} but it's not in correct format, please convert it into following format:\n{format_instructions}\n\nIf the response doesn't seem to be relevant to the expected format instructions, return 'N/A'")
            register_prompt_template("retry_parse_template", retry_parse_template)
            prompt_llm=llm or GlobalSettings.get_current_settings().default_llm
            retryChain = LLMChain(llm=prompt_llm, prompt=retry_parse_template)
            format_instructions = prompt_template.output_parser.get_format_instructions()
            if not format_instructions:
                raise Exception(f"Failed to get format instructions for {full_name} from output parser {prompt_template.output_parser}.")
            call_kwargs = {"original_prompt":original_prompt, "original":exception.original, "format_instructions":format_instructions}
            return retryChain, call_kwargs
        
        def process_results(llmChain, result_data, result, is_function_call):
            log_results(result_data, result, is_function_call, verbose, prompt_type)
            if llmChain.prompt.output_parser:    
                if isinstance(llmChain.prompt.output_parser, OpenAIFunctionsPydanticOutputParser):
                    # there is no result probably, but if there is, we ignore it... we are interested only in tha data in function_call_info
                    result = llmChain.prompt.output_parser.parse(result_data["function_call_info"]["arguments"])
                    result_data.pop("function_call_info") # we don't need it anymore, and later in the code we check it its here to create OutputWithFunctionCall
                    result_data.pop("function")

                else:
                    if result:
                        result = llmChain.prompt.output_parser.parse(result)
            return result

        if not is_async:

            @wraps(func)
            def wrapper(*args, **kwargs):
                
                print_log(log_object=f"> Entering {name} prompt decorator chain", log_level=prompt_type.log_level if prompt_type else logging.DEBUG,color=LogColors.WHITE_BOLD)
                llmChain, chain_args = prepare_call_args(*args, **kwargs)

                try:
                    result_data = llmChain(**chain_args)

                    result = result_data[llmChain.output_key]
                    is_function_call = result_data.get("function_call_info")
                    result = process_results(llmChain, result_data, result,is_function_call)
                        
                    
                except OutputParserException as e:
                    if retry_on_output_parsing_error and isinstance(e, OutputParserExceptionWithOriginal):
                        prompt_template = llmChain.prompt 
                        retryChain, call_kwargs = get_retry_parse_call_args(prompt_template, e, lambda: llmChain.prompt.format(**chain_args))
                        result = retryChain.predict(**call_kwargs)
                        if verbose or prompt_type:
                            print_log(log_object=f"\nResult:\n{result}", log_level=prompt_type.log_level if not verbose else 100,color=prompt_type.color if prompt_type else LogColors.BLUE)
                        parsed = prompt_template.output_parser.parse(result)
                        return parsed
                    else: 
                        raise e

                print_log(log_object=f"> Finished chain", log_level=prompt_type.log_level if prompt_type else logging.DEBUG,color=LogColors.WHITE_BOLD)
                if "function_call_info" in result_data:
                    return _generate_output_with_function_call(result=result, result_data=result_data, verbose=verbose,callbacks=kwargs.get("callbacks"))
                return result

            

            
            return wrapper
        
        else:
            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                
                
                print_log(log_object=f"> Entering {name} prompt decorator chain", log_level=prompt_type.log_level if prompt_type else logging.DEBUG,color=LogColors.WHITE_BOLD)
                llmChain, chain_args = prepare_call_args(*args, **kwargs)
                
                try:
                    
                    result_data = await llmChain.acall(**chain_args)
                    result = result_data[llmChain.output_key]
                    is_function_call = result_data.get("function_call_info")
                    result = process_results(llmChain, result_data, result,is_function_call)
                    
                    
                except OutputParserException as e:
                    if retry_on_output_parsing_error and isinstance(e, OutputParserExceptionWithOriginal):
                        prompt_template = llmChain.prompt 
                        retryChain, call_kwargs = get_retry_parse_call_args(prompt_template, e, lambda: llmChain.prompt.format(**chain_args))
                        result = await retryChain.apredict(**call_kwargs)
                        if verbose or prompt_type:
                            print_log(log_object=f"\nResult:\n{result}", log_level=prompt_type.log_level if not verbose else 100,color=prompt_type.color if prompt_type else LogColors.BLUE)
                        parsed = prompt_template.output_parser.parse(result)
                        return parsed
                    else: 
                        raise e

                print_log(log_object=f"> Finished chain", log_level=prompt_type.log_level if prompt_type else logging.DEBUG,color=LogColors.WHITE_BOLD)
                if "function_call_info" in result_data:
                    return _generate_output_with_function_call(result=result, result_data=result_data, verbose=verbose,callbacks=kwargs.get("callbacks"))
                return result
            return async_wrapper
    if func:
        return decorator(func)
    else:
        return decorator




def _generate_output_with_function_call(result:Any, result_data:dict, verbose, callbacks):
    """ get parsed result, function call data from llm and list of functions and build  OutputWithFunctionCall """
    # find the function first:
    
    _function = result_data["function"]
    if result_data.get("function_call_info"):
        _tool_arguments = result_data["function_call_info"]["arguments"]
        if isinstance(_function, BaseTool):
            # langchain hack >> "__arg1" as a single argument hack
            _is_single_arg_hack="__arg1" in _tool_arguments and len(_tool_arguments)==1
            tool_input= _tool_arguments["__arg1"] if _is_single_arg_hack else _tool_arguments
            _tool_arguments = tool_input
            def _sync_function(arguments=tool_input):
                return _function.run(tool_input=arguments, verbose=verbose, callbacks=callbacks)
            
            async def _async_function(arguments=tool_input):
                return await _function.arun(tool_input=arguments, verbose=verbose, callbacks=callbacks)
                
            

        elif callable(_function):
            # TODO: add support for verbose and callbacks
            
            is_async = inspect.iscoroutinefunction(_function)
            
            if is_async:
                _async_function = _function
                _sync_function = None
            else:
                _sync_function = _function
                _async_function = None
        else:
            raise TypeError(f"Invalid function type: {_function} of type {type(_function)}")

        return OutputWithFunctionCall(
                output=result,
                output_text=result_data["text"],
                output_message=result_data["message"],
                function=_sync_function,
                function_async=_async_function,
                function_name=result_data["function_call_info"]["name"],
                function_args=result_data["function_call_info"]["arguments"],
                function_arguments=_tool_arguments
            )
    else:
        return OutputWithFunctionCall(
                output=result,
                output_message=result_data["message"],
                output_text=result_data["text"],
            )

def log_results(result_data, result, is_function_call, verbose, prompt_type=None):
    if verbose or prompt_type:
        print_log(log_object=f"\nResult:\n{result}", log_level=prompt_type.log_level if verbose else 100,color=prompt_type.color if prompt_type else LogColors.BLUE)
        if is_function_call:
            function_call_info_str = json.dumps(result_data.get('function_call_info'),indent=4)
            print_log(log_object=f"\nFunction call:\n{function_call_info_str}", log_level=prompt_type.log_level if verbose else 100,color=prompt_type.color if prompt_type else LogColors.BLUE)



from typing import Type
from pydantic import BaseModel, ValidationError
from pydantic.fields import ModelField



def get_field_type(field_info: ModelField):
    _item_type=None
                
    if field_info.type_==field_info.outer_type_:
        _type=field_info.type_
    elif list == getattr(field_info.outer_type_, '__origin__', None):
        #is list
        _type = list
        
    elif dict == getattr(field_info.outer_type_, '__origin__', None):
        _type=dict
    else:
        raise Exception(f"Unknown type: {field_info.annotation}")
   
    return _type



def is_field_nullable(field_info: ModelField):
    _nullable=field_info.allow_none
    

def get_field_item_type(field_info: ModelField):
    
    if list == getattr(field_info.outer_type_, '__origin__', None):
        return field_info.outer_type_.__args__[0]
    
def align_fields_with_model( data:dict, model:Type[BaseModel]) -> dict:
    res = {}
    data_with_compressed_keys=None
    for field,field_info in model.__fields__.items():
        value=None
        if field in data:
            value=data[field]
        elif field_info.field_info.title is not None:
            if field_info.field_info.title in data:
                value=data[field_info.field_info.title] 
            elif field_info.field_info.title.lower() in data:
                value=data[field_info.field_info.title.lower()]
        elif field_info.field_info.alias:
            if field_info.field_info.alias in data:
                value=data[field_info.field_info.alias]
            elif field_info.field_info.alias.lower() in data:
                value=data[field_info.field_info.alias.lower()]
        else:
            if not data_with_compressed_keys:
                data_with_compressed_keys= {k.lower().replace(" ",""):v for k,v in data.items()}
            compressed_key = field.lower().replace(" ","").replace("_","")
            if compressed_key in data_with_compressed_keys:
                value=data_with_compressed_keys[compressed_key]
        
        if isinstance(value, dict):
            if field_info.type_ and issubclass(field_info.type_, BaseModel):
                value = align_fields_with_model(value, field_info.type_)
        elif isinstance(value, list):
            value = [align_fields_with_model(item, field_info.type_) for item in value]
        res[field]=value
    return res
    

def humanize_pydantic_validation_error(validation_error:ValidationError):
     return "\n".join([ f'{".".join([str(i) for i in err.get("loc")])} - {err.get("msg")} ' for err in  validation_error.errors()])


def sanitize_pydantic_schema(schema:dict):
    """ Pydantic schema uses '$ref' references to definitions, which doesn't go well with LLMs... les fix that  """
    if schema.get("definitions"):
        definitions=schema.pop("definitions")
        for def_key, definition in definitions.items():
            if "title" in definition:
                definition.pop("title") # no need for this
            nested_ref = next((1 for val in definition.get("properties",{}) if isinstance(val,dict) and val.get("$ref") ),None)
            if nested_ref:
                raise Exception(f"Nested $ref not supported! ... probably recursive schema: {def_key}")
        def replace_refs_recursive(schema:dict):
            if isinstance(schema,dict):
                if schema.get("properties"):
                    for k,v in schema.get("properties").items():
                        if v.get("$ref"):
                            schema["properties"][k]=definitions[v["$ref"].split("/")[-1]]
                        elif v.get("properties"):
                            replace_refs_recursive(v)
                        elif v.get("items"):
                            if isinstance(v["items"],dict):
                                ref = v["items"].get("$ref")
                                if  ref:
                                    v["items"]=definitions[ref.split("/")[-1]]

        replace_refs_recursive(schema)
    return schema




import logging
import re
import inspect

from string import Formatter

from typing import Any, Callable, Coroutine, Dict, List, Optional, Union

from pydantic import BaseModel

from textwrap import dedent

from langchain import PromptTemplate
from langchain.prompts import StringPromptTemplate
from langchain.prompts.chat import  MessagesPlaceholder, ChatMessagePromptTemplate, ChatPromptTemplate, ChatPromptValue
from langchain.schema import PromptValue, BaseOutputParser

from promptwatch import register_prompt_template
from .schema import OutputWithFunctionCall
from .common import LogColors, PromptTypeSettings, get_func_return_type, get_function_docs, get_function_full_name, print_log
from .output_parsers import *


def parse_prompts_from_docs(docs:str):
    prompts = []
    for i, prompt_block in enumerate(re.finditer(r"```[^\S\n]*<prompt(?P<role>:\w+)?>\n(?P<prompt>.*?)\n```[ |\t]*\n", docs, re.MULTILINE | re.DOTALL)):
        role = prompt_block.group("role")
        prompt = prompt_block.group("prompt")
        # remove \ escape before ```
        prompt = re.sub(r"((?<=\s)\\(?=```))|^\\(?=```)", "",prompt, flags=re.MULTILINE )
        prompt.strip()
        if not role:
            if i>1:
                raise ValueError("Only one prompt can be defined in code block. If you intend to define messages, you need to specify a role.\nExample:\n```<prompt:role>\nFoo {bar}\n```")
            else:
                prompts.append(prompt)
        else:
            prompts.append((role[1:], prompt))
    if not prompts:
        # the whole document is a prompt
        prompts.append(docs.strip())

    return prompts


class PromptTemplateDraft(BaseModel):
    role:str=None
    input_variables:List[str]
    template:str
    partial_variables_builders:Optional[Dict[str, Callable[[dict], str]]]

    def finalize_template(self, input_variable_values:dict)->Union[MessagesPlaceholder, ChatMessagePromptTemplate, StringPromptTemplate]:
        if self.role=="placeholder":
            return MessagesPlaceholder(variable_name=self.input_variables[0])
        else:
            final_template_string = self.template
            if self.partial_variables_builders:
                
                for final_partial_key, partial_builder in self.partial_variables_builders.items():
                    final_partial_value = partial_builder(input_variable_values)
                    final_template_string=final_template_string.replace(f"{{{final_partial_key}}}",final_partial_value)
        
            final_template_string=final_template_string.strip()
            content_template= PromptTemplate.from_template(final_template_string)
            if self.role:
                return ChatMessagePromptTemplate(role=self.role,prompt=content_template)
            else:
                return content_template
            

def build_template_drafts(template:str, format:str, role:str=None )->PromptTemplateDraft:
    partials_with_params={}
    
    if role !="placeholder" and format=="f-string-extra":
        optional_blocks_regex = list(re.finditer(r"\{\?(?P<optional_partial>.+?)(?=\?\})\?\}", template, re.MULTILINE | re.DOTALL))
        for optional_block in optional_blocks_regex:
            optional_partial = optional_block.group("optional_partial")
            partial_input_variables = {v for _, v, _, _ in Formatter().parse(optional_partial) if v is not None}
            
            if not partial_input_variables:
                raise ValueError(f"Optional partial {optional_partial} does not contain any optional variables.")
            
            
            # replace  {} with [] and all other non-word characters with underscore
            partial_name = re.sub(r"[^\w\[\]]+", "_", optional_partial.replace("{","[").replace("}","]"))
            

            partials_with_params[partial_name] = (optional_partial, partial_input_variables)
            # replace optional partial with a placeholder
            template = template.replace(optional_block.group(0), f"{{{partial_name}}}")

        partial_builders = {} # partial_name: a function that takes in a dict of variables and returns a string...
        
        
                
        for partial_name, (partial, partial_input_variables) in partials_with_params.items():
            # create function that will render the partial if all the input variables are present. Otherwise, it will return an empty string... 
            # it needs to be unique for each partial, since we check only for the variables that are present in the partial
            def partial_formatter(inputs, _partial=partial):
                """ This will render the partial if all the input variables are present. Otherwise, it will return an empty string."""
                missing_param = next((param for param in partial_input_variables if param not in inputs or not inputs[param]), None)
                if missing_param:
                    return ""
                else:
                    return _partial
            
            partial_builders[partial_name] = partial_formatter

    input_variables = [v for _, v, _, _ in Formatter().parse(template) if v is not None and v not in partials_with_params]
    for partial_name, (partial, partial_input_variables) in partials_with_params.items():
        input_variables.extend(partial_input_variables)

    input_variables=list(set(input_variables))

    if not partials_with_params:
        partials_with_params=None
        partial_builders=None
    if not role:
        return PromptTemplateDraft(input_variables=input_variables, template=template, partial_variables_builders=partial_builders)
    elif role=="placeholder":
        if len(input_variables)>1:
            raise ValueError(f"Placeholder prompt can only have one input variable, got {input_variables}")
        elif len(input_variables)==0:
            raise ValueError(f"Placeholder prompt must have one input variable, got none.")
        return PromptTemplateDraft(template=template,  input_variables=input_variables,  partial_variables_builders=partial_builders, role="placeholder")
    else:
        return PromptTemplateDraft(role=role, input_variables=input_variables, template=template, partial_variables_builders=partial_builders)
        

class PromptDecoratorTemplate(StringPromptTemplate):
    template_string:str
    prompt_template_drafts:Union[PromptTemplateDraft, List[PromptTemplateDraft]]
    template_name:str
    template_format:str
    optional_variables:List[str]
    optional_variables_none_behavior:str
    default_values:Dict[str,Any]
    format_instructions_parameter_key:str
    template_version:str=None
    prompt_type:PromptTypeSettings = None

    

    

    @classmethod 
    def build(cls, 
              template_string:str, 
              template_name:str,
              template_format:str="f-string-extra", 
              output_parser:Union[None, BaseOutputParser]=None, 
              optional_variables:Optional[List[str]]=None,
              optional_variables_none_behavior:str="skip_line", 
              default_values:Optional[Dict[str,Any]]=None,
              format_instructions_parameter_key:str="FORMAT_INSTRUCTIONS",
              template_version:str=None,
              prompt_type:PromptTypeSettings = None
            )->"PromptDecoratorTemplate":
            
        if template_format not in ["f-string","f-string-extra"]:
            raise ValueError(f"template_format must be one of [f-string, f-string-extra], got {template_format}")
        

        prompts = parse_prompts_from_docs(template_string)
        
        if isinstance(prompts,list):
            prompt_template_drafts=[]
            input_variables=[]
            
        for prompt in prompts:
            if isinstance(prompt,str):
                prompt_template_drafts=build_template_drafts(prompt, format=template_format)
                input_variables=prompt_template_drafts.input_variables
                #there should be only one prompt if it's a string
                break
            else:
                (role, content_template)= prompt
                message_template = build_template_drafts(content_template, format=template_format, role=role)
                input_variables.extend(message_template.input_variables)
                prompt_template_drafts.append(message_template)
        



        return cls(
                input_variables=input_variables, #defined in base
                output_parser=output_parser,#defined in base
                prompt_template_drafts=prompt_template_drafts,
                template_name=template_name,
                template_version=template_version,
                template_string=template_string,
                template_format=template_format,
                optional_variables=optional_variables,
                optional_variables_none_behavior=optional_variables_none_behavior,
                default_values=default_values,
                format_instructions_parameter_key=format_instructions_parameter_key,
                prompt_type=prompt_type
            )
        
    @classmethod 
    def from_func(cls, 
                  func:Union[Callable, Coroutine], 
                  template_name:str=None, 
                  template_version:str=None, 
                  output_parser:Union[str,None, BaseOutputParser]="auto", 
                  template_format:str = "f-string-extra",
                  format_instructions_parameter_key:str="FORMAT_INSTRUCTIONS",
                  prompt_type:PromptTypeSettings = None
                  )->"PromptDecoratorTemplate":
        
        template_string = get_function_docs(func)  
        template_name=template_name or get_function_full_name(func)
        return_type = get_func_return_type(func)
      
        if output_parser=="auto":
            if return_type==str or return_type==None:
                output_parser = "str"
            elif return_type==dict:
                output_parser = "json"
            elif return_type==list:
                output_parser = "list"
            elif return_type==bool:
                output_parser = "boolean"
            elif issubclass(return_type, OutputWithFunctionCall):
                return_type = "str"
            elif issubclass(return_type,BaseModel):
                output_parser = PydanticOutputParser(model=return_type)
            else:
                raise Exception(f"Unsupported return type {return_type}")
        if isinstance(output_parser,str):
            if output_parser=="str":
                output_parser = None
            elif output_parser=="json":
                output_parser = JsonOutputParser()
            elif output_parser=="boolean":
                output_parser = BooleanOutputParser()
            elif output_parser=="markdown":
                if return_type and return_type!=dict:
                    raise Exception(f"Conflicting output parsing instructions. Markdown output parser only supports return type dict, got {return_type}.")
                else:
                    output_parser = MarkdownStructureParser()
            elif output_parser=="list":
                output_parser = ListOutputParser()
            elif  output_parser == "pydantic":
                if issubclass(return_type,BaseModel):
                    output_parser = PydanticOutputParser(model=return_type)
                elif return_type==None:
                    raise Exception(f"You must annotate the return type for pydantic output parser, so that we can infer the model")
                else:
                    raise Exception(f"Unsupported return type {return_type} for pydantic output parser")
            elif output_parser=="functions":
                if not return_type:
                    raise Exception(f"You must annotate the return type for functions output parser, so that we can infer the model")
                elif not issubclass(return_type,OutputWithFunctionCall):
                    if issubclass(return_type,BaseModel):
                        output_parser = OpenAIFunctionsPydanticOutputParser(model=return_type)
                    else:
                        raise Exception(f"Functions output parser only supports return type pydantic models, got {return_type}")
                else:
                    output_parser=None
            else:
                raise Exception(f"Unsupported output parser {output_parser}")

        
        default_values = {k:v.default for k,v in inspect.signature(func).parameters.items() if v.default!=inspect.Parameter.empty}

        return cls.build(
            template_string=template_string,
            template_name=template_name,
            template_version=template_version,
            output_parser=output_parser,
            template_format=template_format,
            optional_variables=[*default_values.keys()],
            default_values=default_values,
            format_instructions_parameter_key=format_instructions_parameter_key,
            prompt_type=prompt_type
        )


    def get_final_template(self, **kwargs: Any)->PromptTemplate:
        """Create Chat Messages."""
        
        
        if self.default_values:
            # if we have default values, we will use them to fill in missing values
            kwargs = {**self.default_values, **kwargs}

        kwargs = {k:v for k,v in kwargs.items() if v is not None}

        if isinstance(self.prompt_template_drafts,list):
            message_templates=[]
            for message_draft in self.prompt_template_drafts:
                msg_template = message_draft.finalize_template(kwargs)
                
                message_templates.append(msg_template)

            template = ChatPromptTemplate(messages=message_templates, output_parser=self.output_parser)
        else:
            template = self.prompt_template_drafts.finalize_template(kwargs)
            template.output_parser = self.output_parser
            
            
        register_prompt_template(self.template_name, template, self.template_version)
        return template

    def format_prompt(self, **kwargs: Any) -> PromptValue:
        if self.format_instructions_parameter_key in self.input_variables and  not kwargs.get(self.format_instructions_parameter_key)  and self.output_parser :
            # add format instructions to inputs
            kwargs[self.format_instructions_parameter_key] = self.output_parser.get_format_instructions()
            
        final_template = self.get_final_template(**kwargs)
        kwargs = {k:(v if v is not None else "" ) for k,v in  kwargs.items() if k in  final_template.input_variables}
        if isinstance(final_template,ChatPromptTemplate):

            for msg in list(final_template.messages):
                if isinstance(msg,MessagesPlaceholder):
                    if not kwargs.get(msg.variable_name):
                        kwargs[msg.variable_name] = []
                    
        formatted =  final_template.format_prompt(**kwargs)
        if isinstance(formatted,ChatPromptValue):
            for msg in list(formatted.messages):
                if not msg.content or not msg.content.strip():
                    formatted.messages.remove(msg)
        self.on_prompt_formatted(formatted.to_string())

        return formatted
    
    
    def format(self, **kwargs: Any) ->str:
        formatted = self.get_final_template(**kwargs).format(**kwargs)
        self.on_prompt_formatted(formatted)
        return formatted

    def on_prompt_formatted(self, formatted:str):
        if not self.prompt_type :
            log_level = logging.DEBUG
        else:
            log_level = self.prompt_type.log_level
            
        log_color =  LogColors.DARK_GRAY # we dont want to color the prompt, is's misleading... we color only the output
        print_log(f"Prompt:\n{formatted}",  log_level , log_color)




from email.mime import message
import json
import logging
from multiprocessing import Value
from pyexpat.errors import messages
from typing import Any, Callable, Dict, List, Optional, Union
from langchain import LLMChain
from langchain.schema import LLMResult
from langchain.callbacks.manager import CallbackManagerForChainRun, AsyncCallbackManagerForChainRun
from langchain.tools.base import BaseTool
from langchain.chat_models import ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.prompts.chat import  ChatPromptValue
from langchain.schema import PromptValue, ChatGeneration
from pydantic import root_validator

from .common import LlmSelector

from .function_decorator import get_function_schema
try:
    from langchain.tools.convert_to_openai import format_tool_to_openai_function
except ImportError:
    pass

MODELS_WITH_FUNCTIONS_SUPPORT=["gpt-3.5-turbo-0613","gpt-4-0613"]



class LLMDecoratorChain(LLMChain):

    llm_selector:LlmSelector=None
    """ Optional LLM selector to pick the right LLM for the job. """
    capture_stream:bool=False
    expected_gen_tokens:Optional[int]=None

    def select_llm(self, prompts):
        if self.llm_selector:
            # we pick the right LLM based on the first prompt
            first_prompt = prompts[0]
            if isinstance(first_prompt, ChatPromptValue):
                llm = self.llm_selector.get_llm(first_prompt.messages,**self._additional_llm_selector_args())
            else:
                llm =  self.llm_selector.get_llm(first_prompt.to_string(),**self._additional_llm_selector_args())
        else:
            llm = self.llm
        return llm
    
    def _additional_llm_selector_args(self):
        return {
            "expected_generated_tokens":self.expected_gen_tokens, 
            "streaming":self.capture_stream
            }

    def generate(
        self,
        input_list: List[Dict[str, Any]],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> LLMResult:
        """Generate LLM result from inputs."""
        prompts, stop = self.prep_prompts(input_list, run_manager=run_manager)


        return self.select_llm(prompts).generate_prompt(
            prompts, stop, callbacks=run_manager.get_child() if run_manager else None
        )

    async def agenerate(
        self,
        input_list: List[Dict[str, Any]],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> LLMResult:
        """Generate LLM result from inputs."""
        prompts, stop = await self.aprep_prompts(input_list, run_manager=run_manager)

        return await self.select_llm(prompts).agenerate_prompt(
            prompts, stop, callbacks=run_manager.get_child() if run_manager else None
        )

class LLMDecoratorChainWithFunctionSupport(LLMDecoratorChain):


    functions:Optional[List[Union[Callable, BaseTool]]] = []
    function_schemas:List[dict]=None
    function_call_output_key:str="function_call_info"
    function_output_key:str="function"
    message_output_key:str="message"

    @property
    def output_keys(self) -> List[str]:
        """Will always return text key.

        :meta private:
        """
        return [self.output_key, self.function_output_key, self.function_call_output_key]

    
    @root_validator(pre=True)
    def validate_and_prepare_chain(cls, values):
        functions = values.get("functions",None)
        llm = values.get("llm",None)
        if functions:
            function_schemas=[None for _ in functions]
            for i,f in enumerate(functions):
                if isinstance(f, BaseTool):
                    function_schemas[i] = format_tool_to_openai_function(f)
                elif callable(f) and hasattr(f,"get_function_schema"):
                    schema = get_function_schema(f)
                    if not schema:
                        raise ValueError(f"Invalid item value in functions. Unable to retrieve schema from function {f}")
                    function_schemas[i] =schema
                else:
                    raise ValueError(f"Invalid item value in functions. Only Tools or functions decorated with @llm_function are allowed. Got: {f}")
            values["function_schemas"] = function_schemas
        if not llm:
            raise ValueError("llm must be defined")
        elif functions:
            if not isinstance(llm,ChatOpenAI):
                raise ValueError(f"llm must be a ChatOpenAI instance. Got: {llm}")
            else:
                if llm.model_name not in MODELS_WITH_FUNCTIONS_SUPPORT:
                    # keeping this as a warning to keep it future proof
                    logging.warn(f"WARNING! Model {llm.model_name} likely does not support functions. Functions will be likely ignored!)")

        return values
    


    def _additional_llm_selector_args(self):
        args = super()._additional_llm_selector_args()
        args["function_schemas"]=self.function_schemas
        return args
    
    def preprocess_inputs(self, input_list):
        additional_kwargs={}
        if "function_call" in input_list[0]:
            for input in input_list:
                function_call=input.pop("function_call")
            # function call should be only one... and the same for all inputs... there shouldn't be more anyway
            if not isinstance(function_call,str):
                #find the index of the function in self.functions
                function_index = self.functions.index(function_call)
                if function_index == -1:
                    raise ValueError(f"Invalid function call. Function {function_call} is not defined in this chain")
                function_call = {"name": self.function_schemas[function_index]["name"]}
            elif function_call not in ["none","auto"]:
                function_call = {"name": function_call}

            additional_kwargs["function_call"]=function_call 
        return additional_kwargs



    def generate(
        self,
        input_list: List[Dict[str, Any]],
        run_manager: Optional[CallbackManagerForChainRun] = None,
    ) -> LLMResult:
        """Generate LLM result from inputs."""
        
        additional_kwargs = self.preprocess_inputs(input_list)
           
        prompts, stop = self.prep_prompts(input_list, run_manager=run_manager)
        if self.functions:
            
            chat_model:BaseChatModel=self.select_llm(prompts)
          
            messages = [prompt.to_messages() for prompt in prompts]

            result =  chat_model.generate(messages=messages, 
                                        stop=stop, callbacks=run_manager.get_child() if run_manager else None,
                                        functions=self.function_schemas,
                                        **additional_kwargs
                                        )
            return result
        else:
            return self.select_llm(prompts).generate_prompt(
                prompts, stop, callbacks=run_manager.get_child() if run_manager else None
            )

    async def agenerate(
        self,
        input_list: List[Dict[str, Any]],
        run_manager: Optional[AsyncCallbackManagerForChainRun] = None,
    ) -> LLMResult:
        """Generate LLM result from inputs."""
        additional_kwargs = self.preprocess_inputs(input_list)
    
        prompts, stop = await self.aprep_prompts(input_list, run_manager=run_manager)
        if self.functions:
            chat_model:BaseChatModel=self.select_llm(prompts)
            if len(prompts)!=1:
                raise ValueError("Only one prompt is supported when using functions")
            messages = [prompt.to_messages() for prompt in prompts]
             
            return  await chat_model.agenerate(messages=messages, 
                                         stop=stop, callbacks=run_manager.get_child() if run_manager else None,
                                         functions=self.function_schemas,
                                         **additional_kwargs
                                         )
        else:
            return await self.select_llm(prompts).agenerate_prompt(
                prompts, stop, callbacks=run_manager.get_child() if run_manager else None
            )
        
    def _create_output(self,generation):
        res = {
                self.output_key: generation.text,
                self.function_call_output_key: None,
                self.function_output_key: None,
             }
        if isinstance(generation, ChatGeneration):
            res[self.message_output_key] = generation.message
            # let's make a copy of the function call so that we don't modify the original
            function_call = dict(generation.message.additional_kwargs.get("function_call")) if generation.message.additional_kwargs else {}
            if function_call:
                if isinstance(function_call["arguments"],str):
                    function_call["arguments"]=json.loads(function_call["arguments"])
            if generation.message.additional_kwargs and generation.message.additional_kwargs.get("function_call"):
                res[self.function_call_output_key] = function_call
                res[self.function_output_key] = self.find_func(function_call["name"]) if function_call else None
        return res

    def find_func(self,function_name):
        for i,f in enumerate(self.function_schemas):
            if f["name"] == function_name:
                return self.functions[i]
        #else (not found)
        ## TODO: raise error or retry?
    

    def create_outputs(self, response: LLMResult) -> List[Dict[str, str]]:
        """Create outputs from response."""
        
        return [
            self._create_output(generation[0])
            for generation in response.generations
        ]


    


from .common import LogColors, GlobalSettings, print_log, PromptTypes, PromptTypeSettings, LlmSelector
from .schema import OutputWithFunctionCall
from .prompt_decorator import PromptDecoratorTemplate
from .streaming_context import StreamingContext
from .prompt_decorator import llm_prompt
from .function_decorator import llm_function, get_function_schema

__version__="0.0.11"





import datetime
import logging
from textwrap import dedent
from typing import Dict, List, Type, TypeVar, Union
from langchain.output_parsers import PydanticOutputParser
from langchain.schema import BaseOutputParser, OutputParserException

import re
import json
import yaml
from pydantic import BaseModel, ValidationError
from pydantic.fields import ModelField
from .function_decorator import llm_function
from .pydantic_helpers import *


class OutputParserExceptionWithOriginal(OutputParserException):
    """Exception raised when an output parser fails to parse the output of an LLM call."""

    def __init__(self, message: str, original: str, original_prompt_needed_on_retry:bool=False) -> None:
        super().__init__(message)
        self.original = original
        self.original_prompt_needed_on_retry=original_prompt_needed_on_retry

    def __str__(self) -> str:
        return f"{super().__str__()}\nOriginal output:\n{self.original}"


class ListOutputParser(BaseOutputParser):
    """Class to parse the output of an LLM call to a list."""

    @property
    def _type(self) -> str:
        return "list"

    def parse(self, text: str) -> List[str]:
        """Parse the output of an LLM call."""

        pattern = r"^[ \t]*(?:[\-\*\+]|\d+\.)[ \t]+(.+)$"
        matches = re.findall(pattern, text, flags=re.MULTILINE)
        if not matches and text:
            logging.warning(
                f"{self.__class__.__name__} : LLM returned {text} but we could not parse it into a list")
        return matches

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        return "Return result a s bulleted list."

class BooleanOutputParser(BaseOutputParser):
    """Class to parse the output of an LLM call to a boolean."""
    pattern:str
    
    @property
    def _type(self) -> str:
        return "boolean"
    
    def __init__(self, pattern: str = r"((Yes)|(No))([,|.|!]|$)") -> None:
        super().__init__(pattern=pattern)

    def parse(self, text: str) -> bool:
        """Parse the output of an LLM call."""

        
        
        match = re.search(self.pattern, text, flags=re.MULTILINE | re.IGNORECASE)
        
        if not match:
            raise OutputParserExceptionWithOriginal(message=self.get_format_instructions(),original=text, original_prompt_needed_on_retry=True)
        else:
            return match.group(1).lower() == "yes"
        

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        return "Reply only Yes or No.\nUse this format: Final decision: Yes/No"

class JsonOutputParser(BaseOutputParser):
    """Class to parse the output of an LLM call to a Json."""

    @property
    def _type(self) -> str:
        return "json"

    def parse(self, text: str) -> List[str]:
        try:
            # Greedy search for 1st json candidate.
            match = re.search(r"\{.*\}", text.strip(),
                              re.MULTILINE | re.IGNORECASE | re.DOTALL)
            json_str = ""
            if match:
                json_str = match.group()
            json_dict = json.loads(json_str, strict=False)
            return json_dict

        except (json.JSONDecodeError) as e:

            msg = f"Invalid JSON\n {text}\nGot: {e}"
            raise OutputParserExceptionWithOriginal(msg, text)

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        return "Return result as a valid JSON"


T = TypeVar("T", bound=BaseModel)


class PydanticOutputParser(BaseOutputParser[T]):
    """Class to parse the output of an LLM call to a pydantic object."""
    model: Type[T]
    instructions_as_json_example: bool = True

    def __init__(self, model: Type[T], instructions_as_json_example: bool = True):
        super().__init__(model=model, instructions_as_json_example=instructions_as_json_example)

    @property
    def _type(self) -> str:
        return "pydantic"

    def parse(self, text: str) -> T:
        try:
            # Greedy search for 1st json candidate.
            match = re.search(r"\{.*\}", text.strip(),re.MULTILINE | re.IGNORECASE | re.DOTALL)
            json_str = ""
            if match:
                json_str = match.group()
            json_dict = json.loads(json_str, strict=False)

            return self.model.parse_obj(json_dict)

        except (json.JSONDecodeError) as e:
            msg = f"Invalid JSON\n {text}\nGot: {e}"
            raise OutputParserExceptionWithOriginal(msg, text)

        except ValidationError as e:
            try:
                json_dict_aligned = align_fields_with_model(json_dict, self.model)
                return self.model.parse_obj(json_dict_aligned)
            except ValidationError as e:
                err_msg =humanize_pydantic_validation_error(e)
                raise OutputParserExceptionWithOriginal(f"Data are not in correct format: {text}Errors: {err_msg}",text) 
        
    def get_json_example_description(self, model:Type[BaseModel], indentation_level=0):
        field_descriptions = {}
        for field, field_info in model.__fields__.items():

            _item_type = None

            if field_info.type_ == field_info.outer_type_:
                _type = field_info.type_
            elif list == getattr(field_info.outer_type_, '__origin__', None):
                # is list
                _type = list
                _item_type = field_info.outer_type_.__args__[0]
            elif dict == getattr(field_info.outer_type_, '__origin__', None):
                _type = dict
            else:
                raise Exception(f"Unknown type: {field_info.annotation}")
            _nullable = field_info.allow_none
            _description = field_info.field_info.description

            if issubclass(_type, BaseModel):
                field_descriptions[field] = (
                    self.get_json_example_description(_type, indentation_level+1))
            elif _type == str:
                desc = _get_str_field_description(field_info)
                field_descriptions[field] = (desc)
            elif _type == datetime:
                field_descriptions[field] = (
                    "an ISO formatted datetime string")
            elif _type == list:
                list_desc = f"[ {_description} ... list of {_item_type} ]"
                field_descriptions[field] = (list_desc)
            elif _type == dict:
                dict_desc = f"{{ ... {_description} ... }}"
                field_descriptions[field] = (dict_desc)
            elif _type == int:
                field_descriptions[field] = ("an integer")
            elif _type == float:
                field_descriptions[field] = ("a float number")
            else:
                field_descriptions[field] = "?"            

        lines = []
        for field, field_info in model.__fields__.items():
            desc_lines = "\n".join(
                ("\t"*indentation_level+line for line in field_descriptions[field].splitlines())).strip()

            lines.append("\t"*indentation_level + f"\"{field}\": {desc_lines}")

        return "{\n" + ",\n".join(lines) + "\n}"

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        if not self.instructions_as_json_example:
            return "Return result as a valid JSON that matched this json schema definition:\n" + yaml.safe_dump(self.model.schema())
        else:

            return dedent(f"""```json\n{self.get_json_example_description(self.model)}\n```""").strip()

class OpenAIFunctionsPydanticOutputParser(BaseOutputParser[T]):
    model: Type[T]

    @property
    def _type(self) -> str:
        return "opanai_functions_pydantic"
    
    def __init__(self, model: Type[T]):
        super().__init__(model=model)

    def parse(self, function_call_arguments:dict ) -> T:
        try:
            return self.model.parse_obj(function_call_arguments)
        except ValidationError as e:
            err_msg =humanize_pydantic_validation_error(e)
            serialized= json.dumps(function_call_arguments)
            raise OutputParserExceptionWithOriginal(f"Function call arguments are not in correct format: {serialized}Errors: {err_msg}",serialized)
        

    def get_format_instructions(self) -> str:
        return "" # will be handled by openai
    
    
    def build_llm_function(self):
        @llm_function()
        def generate_response( output:self.model) -> T:
            """ Use this to transform the data into desired format. """
            #above is a description for LLM...
            return output
        return generate_response


class CheckListParser(ListOutputParser):
    """Parses list a a dictionary... assume this format:
        - KeyParma1: Value1
        - KeyPara2: Value2
        ...
    """

    def __init__(self, model: Type[T] = None):
        self.model = model

    @property
    def _type(self) -> str:
        return "checklist"

    def get_instructions_for_model(self, model: Type[T]) -> str:
        fields_bullets = []
        for field in model.__fields__.values():
            description = [field.field_info.description]
            if field.field_info.extra.get("one_of"):
                description += "one of these values: [ "
                description += " | ".join(field.field_info.extra.get("one_of"))
                description += " ]"
            if field.field_info.extra.get("example"):
                description += f"e.g. {field.field_info.extra.get('example')}"
            if description:
                description = " ".join(description)
            else:
                description = "?"
            fields_bullets.append(f"- {field.name}: {description}")

    def parse(self, text: str) -> Union[dict, T]:
        """Parse the output of an LLM call."""

        pattern = r"^[ \t]*(?:[\-\*\+]|\d+\.)[ \t]+(.+)$"
        matches = re.findall(pattern, text, flags=re.MULTILINE)
        result = {}
        if not matches:
            raise OutputParserExceptionWithOriginal(message="No matches found", original_output=text)
        for match in matches:
            key, value = match.split(":", 1)
            result[key.strip()] = value.strip()

        return matches

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        res = "Return result a s bulleted list in this format:\n"
        if self.model:
            res += self.get_instructions_for_model(self.model)
        else:
            res += "\n- Key1: Value1\n- Key2: Value2\n- ..."


class MarkdownStructureParser(ListOutputParser):
    model: Type[T] = None
    level: int = 1
    sections_parsers: Dict[str, Union[BaseOutputParser, dict]] = None

    def __init__(self,  model: Type[T] = None, sections_parsers: Dict[str, Union[dict, BaseOutputParser]] = None, level=1):

        super().__init__(model=model, sections_parsers=sections_parsers, level=level)
        if model:
            for field, field_info in model.__fields__.items():
                if sections_parsers and field in self.sections_parsers:
                    # if section parser was already provided, skip
                    if not type(self.sections_parsers.get(field)) == dict:
                        continue
                field_type = get_field_type(field_info)
                if get_field_type(field_info) == list:
                    item_type = get_field_item_type(field_info)
                    if item_type == str or item_type is None:
                        self.sections_parsers[field] = ListOutputParser()
                    else:
                        raise ValueError(
                            f"Unsupported item type {item_type} for property {model}.{field}. Only list of strings is supported.")
                elif field_type == dict:
                    self.sections_parsers[field] = CheckListParser()
                elif field_type and issubclass(field_type, BaseModel):

                    all_sub_str = all(True for sub_field_info in field_type.__fields__.values(
                    ) if get_field_type(sub_field_info) == str)

                    if all_sub_str:

                        self.sections_parsers[field] = MarkdownStructureParser(
                                model=field_type, sections_parsers=sections_parsers.get(field), level=level+1
                            )
                    else:
                        self.sections_parsers[field] = PydanticOutputParser(
                                model=field_type
                            )

                elif field_type == str:

                    self.sections_parsers[field] = None
                else:
                    raise ValueError(
                        f"Unsupported type {field_type} for property {field}.")
        elif sections_parsers:
            for property, property_parser in sections_parsers.items():
                if type(property_parser) == dict:
                    sections_parsers[property] = MarkdownStructureParser(
                        model=None, sections_parsers=property_parser, level=level+1)
                elif type(property_parser) == str:
                    sections_parsers[property] = None
                elif isinstance(property_parser, BaseOutputParser):
                    continue
                else:
                    raise ValueError(
                        f"Unsupported type {model.__fields__[property].annotation} for property {property}. Use a dict or a pydantic model.")
        else:
            self.sections_parsers = {}

    @property
    def _type(self) -> str:
        return "checklist"

    def get_instructions_for_sections(self,  model: Type[T] = None, sections_parsers: Dict[str, BaseOutputParser] = None) -> str:
        section_instructions = []
        if model:
            for field, field_info in model.__fields__.items():
                name: str = field_info.field_info.title or field
                section_instructions.append(self.level*"#" + f" {name}")
                if sections_parsers and sections_parsers.get(field):
                    section_instructions.append(
                        sections_parsers.get(field).get_format_instructions())
                    continue
                else:

                    description = _get_str_field_description(field_info)
                    section_instructions.append(description)
        else:
            for section, parser in sections_parsers.items():
                section_instructions.append(self.level*"#" + f" {section}")
                if isinstance(parser, BaseOutputParser):
                    section_instructions.append(
                        parser.get_format_instructions())
                else:
                    section_instructions.append("?")

        return "\n\n".join(section_instructions)

    def parse(self, text: str) -> List[str]:
        """Parse the output of an LLM call."""

        sections_separators = list(re.finditer(
            r"^#+[ |\t]+(.*)$", text, flags=re.MULTILINE))
        res = {}
        for i, section_separator_match in enumerate(sections_separators):

            section_name = section_separator_match.group(1)
            if self.model:
                section_name = next((field for field, field_info in self.model.__fields__.items() if field_info.field_info.title ==
                                    section_name or field.lower() == section_name.lower() or field_info.alias == section_name), section_name)
            if i < len(sections_separators)-1:
                section_content = text[section_separator_match.end(
                ):sections_separators[i+1].start()]
            else:
                section_content = text[section_separator_match.end():]

            parsed_content = None
            if self.sections_parsers and self.sections_parsers.get(section_name, None) or self.sections_parsers.get(section_separator_match.group(1)):
                parser = self.sections_parsers.get(
                    section_name, None) or self.sections_parsers.get(section_separator_match.group(1))
                if isinstance(parser, BaseOutputParser):
                    parsed_content = parser.parse(section_content)
            if not parsed_content:
                parsed_content = section_content.strip()

            res[section_name] = parsed_content

        if self.model:
            try:
                return self.model(**res)
            except ValidationError as e:
                try:
                    res_aligned = align_fields_with_model(res, self.model)
                    return self.model.parse_obj(res_aligned)
                except ValidationError as e:
                    err_msg =humanize_pydantic_validation_error(e)
                    raise OutputParserExceptionWithOriginal(f"Data are not in correct format: {text}\nGot: {err_msg}",text) 
        else:
            return res

    def get_format_instructions(self) -> str:
        """Instructions on how the LLM output should be formatted."""
        res = "Return result as a markdown in this format:\n"
        if self.model or self.sections_parsers:
            res += self.get_instructions_for_sections(
                self.model, self.sections_parsers)

        else:
            res += "# Section 1\n\ndescription\n\n#Section 2\n\ndescription\n\..."
        return res


def _get_str_field_description(field_info: ModelField, ignore_nullable: bool = False):
    _nullable = field_info.allow_none
    _description = field_info.field_info.description
    _example = field_info.field_info.extra.get("example")
    _enum = field_info.field_info.extra.get("enum")
    _regex = field_info.field_info.extra.get("regex")
    _one_of = _enum or field_info.field_info.extra.get("one_of")
    _regex = field_info.field_info.extra.get("regex")
    description = []
    if _description:
        description.append(_description)
    if _one_of:
        description.append("one of these values: [ ")
        description.append(" | ".join([f"\"{enum_val}\"" for enum_val in _one_of]))
        description.append(" ]")
    if _example:
        description.append(f"e.g. {_example}")
    if _nullable and not ignore_nullable:
        description.append("... or null if not available")
    if _regex and not _enum:
        description.append(f"... must match this regex: {_regex}")

    if description:
        description = " ".join(description)
    else:
        description = "?"

    return (description if _one_of else f"\" {description} \"")



import re
import inspect
from enum import Enum
from functools import wraps
from typing import Callable,  Union, Type
from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.schema import  field_schema
import copy

from .pydantic_helpers import sanitize_pydantic_schema
from .common import  get_arguments_as_pydantic_fields, get_function_docs, get_function_full_name

class DocstringsFormat(Enum):
    AUTO = "auto"
    GOOGLE = "google"
    SPHINX = "sphinx"
    NUMPY = "numpy"

def get_function_schema(func):
    if callable(func) and hasattr(func,"get_function_schema"):
        return func.get_function_schema(func)
    else:
        return None


def llm_function(
        function_name:str=None,
        validate_docstrings:bool=True,
        docstring_format:str="auto",
        arguments_schema:Type[BaseModel]=None,
        **kwargs
        ):
    """
    Decorator for functions that take a language model as first argument.


    Args:
        - function_name: the name of the function for LLM. If not provided, the name of the function will be used

        - argument_schema
            - `auto` (default): the schema is automatically inferred from the function signature. If docstrings are provided, they will be used to enhance function description
            - `pydantic` -  expects a pydantic model as first and only argument ('self' being ignored) - allows for controllable schema
            - `docstring` - parses the schema ONLY from the docstring

            or define pydantic model:
            ```
            @llm_function(argument_schema=MyFunctionPydanticArgsSchema)
            def my_function(self, question:str)->bool:
                ...
            ```

            or define a simple dict with descriptions:
            ```
            @llm_function(argument_schema={"question": "Question to ask"})
            def my_function(self, question:str)->bool:
                ...
            ```

        - validate_docstrings: if True, the docstrings will be parsed and validated against function. If parsing or validation fails, an error will be raised

        - docstring_format: the format of the docstring
            -  `auto` (default): the format is automatically inferred from the docstring
            -  `google`: the docstring is parsed as markdown (see [Google docstring format](https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html))
            -  `numpy`: the docstring is parsed as markdown (see [Numpy docstring format](https://numpydoc.readthedocs.io/en/latest/format.html))
            -  `sphinx`: the docstring is parsed as sphinx format (see [Sphinx docstring format](https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html))

    Examples:
    
        Google style docstrings:
        ```python
        @llm_function
        def function_with_google_docstrings(self, question:str)->bool:
            \"\"\"
            This is a function with google docstrings
            Args:  # but accepts also Parameters, Arguments
                question (str): Question to ask
            \"\"\"
        ```

        ```python
        @llm_function
        def function_with_numpy_docstrings(self, question:str)->bool:
            \"\"\"
            This is a function with google docstrings
            Parameters # but accepts also Args, Arguments
            ----------
                question :int
                    Question to ask
                
            \"\"\"
        ```

        ```python
        @llm_function
        def function_with_sphinx_docstrings(self, question:str)->bool:
            \"\"\"
            This is a function with google docstrings
            :param question: Question to ask
                
            \"\"\"
        ```

    Note: 
        The fact whether the parameter is optional is inferred from the function signature. You can also document it, but it will be just part of the description. If typing annotations are used (i.e. Optional[str]), this will be stripped to native type.
    
    """
    
    if callable(function_name):
        # this is the case when the decorator is called without arguments
        # we initialize params with default values
        func = function_name
        function_name=None
    else:
        func = None


    def get_function_schema(_func, _validate_docstrings=validate_docstrings):
            return build_func_schema(_func, 
                                            function_name = function_name, 
                                            format=docstring_format, 
                                            validate_docstrings=_validate_docstrings,
                                            arguments_schema=arguments_schema
                                          )
    
    def decorator(func):
        
        is_async = inspect.iscoroutinefunction(func)

        if not is_async:
            @wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            
        else:
            @wraps(func)
            async def func_wrapper(*args, **kwargs):
                return await func(*args, **kwargs)
        
        
        
       

        func_wrapper.get_function_schema=get_function_schema
        return func_wrapper
    
    if func:
        return decorator(func)
    else:
        return decorator
    

class LllFunctionWithModifiedSchema:
    def __init__(self, func, modified_schema:dict):
        self.func = func
        self._function_schema = modified_schema


    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def get_function_schema(self):
        return self._function_schema
        
         



def build_func_schema(
        func:Callable, 
        function_name:str=None, 
        format:Union[DocstringsFormat,str]="auto", 
        validate_docstrings:bool=True,
        arguments_schema:Type[BaseModel]=None
        ):
    
    if isinstance(format,str):
        format = DocstringsFormat(format)

    func_docs = get_function_docs(func)  

    if function_name and not re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", function_name):
        raise ValueError(f"Invalid function name: {function_name} for {get_function_full_name(func)}. Only letters, numbers and underscores are allowed. The name must start with a letter or an underscore.")
   
    func_name =  function_name or func.__name__
    args_schema = None
    
    if arguments_schema:
        args_schema = arguments_schema.schema()
        sanitize_pydantic_schema(args_schema)
        args_schema={
            "type":"object",
            "properties":args_schema["properties"],
            "required":args_schema["required"]
        }
    else:

        arguments_fields = get_arguments_as_pydantic_fields(func)

        if func_docs:
            docsctrings_param_description = find_and_parse_params_from_docstrings(func_docs,format=format)
            if docsctrings_param_description:
                if validate_docstrings:
                    documented_params=set(docsctrings_param_description.keys()) 
                    implemented_params= set(arguments_fields.keys())
                    not_implemented = documented_params - implemented_params  # Set difference: keys in set1 but not in set2
                    not_documented = implemented_params - documented_params  # Set difference: keys in set2 but not in set1

                    if not_implemented or not_documented:
                        errs = []
                        if not_implemented:
                            errs.append(f"Missing (not implemented): {not_implemented}")
                        if not_documented:
                            errs.append(f"Missing (not documented): {not_documented}")
                        
                        raise ValueError("Docstrings parameters do not match function signature. "+errs.join(", "))
                
                for arg_name, arg_model_field in arguments_fields.items():
                    arg_docs = docsctrings_param_description.get(arg_name)
                    if arg_docs:
                        
                        arg_model_field.field_info.description =arg_docs["description"]
                        enum = parse_enum_from_docstring_param(arg_docs["type"],arg_docs["description"])
                        if enum:
                            arg_model_field.field_info.extra["enum"] = enum

        args_schema={
            "type":"object",
            "properties":{},
            "required":[]

        }
        for arg_name, arg_model_field in arguments_fields.items():
            param_schema, _ , nested_models_schema = field_schema(arg_model_field, model_name_map={}) 
            
            if nested_models_schema:
                raise NotImplementedError("Nested models are not supported yet")
            
            args_schema["properties"][arg_name] = param_schema
            if arg_model_field.required:
                args_schema["required"].append(arg_name)

    
    def pop_prop_title(schema):
        #title is autogenerated by pydantic and will just costs us tokens....
        if "title" in schema:
            del schema["title"]
        return schema
    args_schema["properties"] = {prop:pop_prop_title(prop_schema) for prop, prop_schema in args_schema["properties"].items()}
    
    description = parse_function_description_from_docstrings(func_docs) if func_docs else None

    if not description:
        raise ValueError(f"LLM Function {get_function_full_name(func)} has no description in docstrings")
    return {
        "name":func_name,
        "description":description,
        "parameters":args_schema
    }
            



        


        
def parse_function_description_from_docstrings(docstring:str)->str:
    # we will return first text until first empty line
    
    lines = docstring.splitlines()
    description = []
    for line in lines:
        line = line.strip()
        if line:
            description.append(line)
        elif description:
            # if we have already some description, we stop at first empty line ... else continue
            break
    return "\n".join(description)





    


    

def find_and_parse_params_from_docstrings(docstring:str,format:DocstringsFormat)->str:
    """
    Find Args section in docstring.
    """

    args_section = None
    args_section_start = 0
    args_section_end = None
    
    if format == DocstringsFormat.AUTO or format == DocstringsFormat.GOOGLE:
        # auto here means more more free format than google
        args_section_start_regex_pattern = r"(^|\n)(Args|Arguments|Parameters)\s*:?\s*\n"
        args_section_end_regex_pattern = r"(^|\n)([A-Z][a-z]+)\s*:?\s*\n"
        if  format == DocstringsFormat.GOOGLE:
            param_start_parser_regex=r"(^|\n)\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*\((?P<type>[^\)]*)\)?\s*:\s*(?=\w+)"
        else:
            param_start_parser_regex=r"(^|\n)\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*\((?P<type>[^\)]*)\)?\s*(-|:)\s*(?=\w+)"
    elif format == DocstringsFormat.NUMPY:
        args_section_start_regex_pattern = r"(^|\n)(Args|Arguments|Parameters)\s*\n\s*---+\s*\n"
        args_section_end_regex_pattern = r"(^|\n)([A-Z][a-z]+)\s*\n\s*---+\s*\n"
        param_start_parser_regex=r"(^|\n)\s*(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)(\s*:\s*(?P<type>[^\)]*)\s*)?\n\s+(?=\w+)"
    elif format == DocstringsFormat.SPHINX:
        args_section_start_regex_pattern = None # we will look for :param everywhere
        args_section_end_regex_pattern=r"(\n)\s*:[a-z]"
        param_start_parser_regex=r"(^|\n)\s*:param\s+(?P<type>[^\)]*)\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)\s*:\s*(?=\w+)"

    if args_section_start_regex_pattern:
        match = re.search(args_section_start_regex_pattern, docstring)
        if match:
            args_section_start = match.end()
            if args_section_end_regex_pattern:
                match = re.search(args_section_end_regex_pattern, docstring[args_section_start:])
                if match:
                    args_section_end = match.start() + args_section_start
            if not args_section_end:
                args_section_end = len(docstring)
            args_section = docstring[args_section_start:args_section_end]
        else:
            args_section=None
    else:
        args_section=docstring
    
    params={}
    if args_section:
        last_param=None
        last_param_end=None
        for param_start_match in re.finditer(param_start_parser_regex, args_section):
            if last_param_end is not None:
                last_param["description"]=args_section[last_param_end:param_start_match.start()].strip()

            param_name = param_start_match.group("name")
            param_type = param_start_match.group("type")
            last_param = {"type":param_type,"description":None}
            last_param_end = param_start_match.end()
            params[param_name]=last_param

        if last_param_end is not None:
            section_end=None
            if args_section_start_regex_pattern is None and args_section_end_regex_pattern:
                # this is handling SPHINX, we didnt parse the start so we cant parse the end until all the params are consumed... now we can parse the end after the last param
                section_end_match= re.search(args_section_end_regex_pattern, docstring[last_param_end:])    
                if section_end_match:
                    section_end=last_param_end + section_end_match.start()
            if not section_end:
                section_end=len(docstring)
            last_param["description"]=args_section[last_param_end:section_end].strip()
        

    if not params and  format == DocstringsFormat.AUTO:
            # try other options            
        options = [DocstringsFormat.NUMPY, DocstringsFormat.SPHINX]
        for option in options:
            result = find_and_parse_params_from_docstrings(docstring,option)
            if result:
                return result
    else:
        return params
        
            
def parse_enum_from_docstring_param(type:str, description)->str:
    enum_pattern=r"\[\s*(?P<value>[\"|'][\w|_|-]+[\"|']\s*(\||\]))+"
    enum_part_match = re.search(enum_pattern, type +" / " +description)
    if not enum_part_match:
        return None
    enum_strings =enum_part_match.group(0).strip("[]").split("|")
    return [enum.strip(" \"'") for enum in enum_strings]



import setuptools

with open("src/langchain_decorators/__init__.py", "rt") as f:
    for line in f.readlines():
        if line.startswith("__version__"):
            __version__ = line.split("=")[1].strip(" \n\"")

setuptools.setup(name='langchain-decorators',
                 version=__version__,
                 description='syntactic sugar for langchain',
                 long_description=open('README.md').read(),
                 long_description_content_type='text/markdown',
                 author='Juraj Bezdek',
                 author_email='juraj.bezdek@blip.solutions',
                 url='https://github.com/ju-bezdek/langchain-decorators',
                 package_dir={"": "src"},
                 packages=setuptools.find_packages(where="src"),
                 license='MIT License',
                 zip_safe=False,
                 keywords='langchain',

                 classifiers=[
                 ],
                 python_requires='>=3.9',
                 install_requires=[
                     "langchain",
                     "promptwatch"
                 ]
                 )



import asyncio
import re
from typing import Any, Callable, Dict, Generic, Optional, TypeVar, Union
from pydantic import BaseModel
from langchain.schema import AIMessage
from langchain.schema import FunctionMessage
import json



T = TypeVar("T")

class OutputWithFunctionCall(Generic[T],BaseModel):
    output_text:str
    output_message:AIMessage
    output:T
    function_name:str =None
    function_arguments:Union[Dict[str,Any],str,None]
    function:Callable = None
    function_async:Callable = None
    result: Any = None
    
    @property
    def is_function_call(self):
        return bool(self.function or self.function_async)
    
    @property
    def support_async(self):
        return bool(self.function_async)
    
    @property
    def support_sync(self):
        return bool(self.function)

    async def execute_async(self):
        """Executes the function asynchronously."""
        if not (self.function or self.function_async):
            raise ValueError("No function to execute")
        if self.function_async:
            if isinstance(self.function_arguments, dict):
                result= await self.function_async(**self.function_arguments)
            else:
                result= await self.function_async(self.function_arguments)
        else:
            await asyncio.sleep(0)
            result= self.function(**self.function_arguments)
            if result and asyncio.iscoroutine(result):
                # this handles special scenario when fake @llm_function is used
                result = await result
        return result
        
    def execute(self):
        """ Executes the function synchronously. 
        If the function is async, it will be executed in a event loop.
        """
        if not (self.function or self.function_async):
            raise ValueError("No function to execute")
        if self.function:
            if isinstance(self.function_arguments, dict):
                result= self.function(**self.function_arguments)
            else:
                result=self.function(self.function_arguments)
        
        else:
            try:
                current_loop = asyncio.get_running_loop()
            except RuntimeError:
                current_loop = None
            if current_loop:
                result= current_loop.run_until_complete(self.function_async(**self.function_arguments))
            else:
                result= asyncio.run(self.function_async(**self.function_arguments))
        self.result = result
        return result
    
    def to_function_message(self, result=None):
        """
        Converts the result to a FunctionMessage... 
        you can override the result collected via execute with your own
        """
        if not result:
            if not self.result:
                raise Exception("The function has not been executed yet, or didn't return a result")
            result = self.result

        if not isinstance(result,str):
            result = json.dumps(result)

        return FunctionMessage(name=self.function_name, content=result)
    



